part of query;

class Foreign {
  /// The table in which foreign key references the column [col]
  final String table;

  /// References column in table [table]
  final String col;

  const Foreign(this.table, this.col);

  static Foreign make(String table, String refCol) =>
      new Foreign(table, refCol);
}

class Unique {
  final bool unique;

  final String group;

  const Unique([this.group]) : unique = true;

  const Unique.Not()
      : unique = false,
        group = null;
}

abstract class CreateColumn<ValType> implements ToSqlable {
  String get colName;

  bool get isNullable;

  bool get isPrimaryKey;

  Foreign get foreignKey;

  Unique get unique;
}

class CreateIntColumn extends CreateColumn<int> {
  bool _nullable = false;

  String _colName;

  bool _autoIncrement = false;

  bool _primary = false;

  Foreign _foreignKey;

  Unique _unique;

  CreateIntColumn(this._colName,
      {bool nullable: false,
      bool autoIncrement: false,
      bool primary: false,
      Foreign foreignKey,
      Unique unique: const Unique.Not()})
      : _nullable = nullable,
        _autoIncrement = autoIncrement,
        _primary = primary,
        _foreignKey = foreignKey,
        _unique = unique;

  CreateIntColumn.Primary(this._colName, {Foreign foreignKey})
      : _primary = true,
        _foreignKey = foreignKey;

  CreateIntColumn.AutoPrimary(this._colName)
      : _primary = true,
        _autoIncrement = true;

  CreateIntColumn nullable() {
    _nullable = true;
    return this;
  }

  CreateIntColumn foreign(String table, {String refCol}) {
    if (_foreignKey != null) throw new Exception("Foreign key alread set!");
    if (refCol == null) refCol = _colName;
    _foreignKey = new Foreign(table, refCol);
    return this;
  }

  String get colName => _colName;

  bool get isNullable => _nullable;

  bool get isPrimaryKey => _primary;

  Foreign get foreignKey => _foreignKey;

  Unique get unique => _unique;

  String toSql() {
    String ret = _colName;
    if (_autoIncrement) {
      ret += ' SERIAL';
    } else {
      ret += ' INT';
    }
    if (!_nullable) ret += ' NOT NULL';
    return ret;
  }
}

class CreateDoubleColumn extends CreateColumn<int> {
  bool _nullable = false;

  String _colName;

  bool _autoIncrement = false;

  bool _primary = false;

  Foreign _foreignKey;

  Unique _unique;

  CreateDoubleColumn(this._colName,
      {bool nullable: false,
      bool autoIncrement: false,
      bool primary: false,
      Foreign foreignKey,
      Unique unique: const Unique.Not()})
      : _nullable = nullable,
        _autoIncrement = autoIncrement,
        _primary = primary,
        _foreignKey = foreignKey,
        _unique = unique;

  CreateDoubleColumn.Primary(this._colName, {Foreign foreignKey})
      : _primary = true,
        _foreignKey = foreignKey;

  CreateDoubleColumn.AutoPrimary(this._colName)
      : _primary = true,
        _autoIncrement = true;

  CreateDoubleColumn nullable() {
    _nullable = true;
    return this;
  }

  CreateDoubleColumn foreign(String table, {String refCol}) {
    if (_foreignKey != null) throw new Exception("Foreign key alread set!");
    if (refCol == null) refCol = _colName;
    _foreignKey = new Foreign(table, refCol);
    return this;
  }

  String get colName => _colName;

  bool get isNullable => _nullable;

  bool get isPrimaryKey => _primary;

  Foreign get foreignKey => _foreignKey;

  Unique get unique => _unique;

  String toSql() {
    String ret = _colName;
    ret += ' double precision';
    if (!_nullable) ret += ' NOT NULL';
    return ret;
  }
}

class CreateBoolColumn extends CreateColumn<bool> {
  bool _nullable = false;

  String _colName;

  bool _primary = false;

  Unique _unique;

  CreateBoolColumn(this._colName,
      {bool nullable: false,
      bool primary: false,
      Unique unique: const Unique.Not()})
      : _nullable = nullable,
        _primary = primary,
        _unique = unique;

  CreateBoolColumn nullable() {
    _nullable = true;
    return this;
  }

  String get colName => _colName;

  bool get isNullable => _nullable;

  bool get isPrimaryKey => _primary;

  Foreign get foreignKey => null;

  Unique get unique => _unique;

  String toSql() => _colName + ' BOOLEAN' + (!_nullable ? ' NOT NULL' : '');
}

class CreateDateTimeColumn extends CreateColumn<bool> {
  bool _nullable = false;

  String _colName;

  bool _primary = false;

  Unique _unique;

  CreateDateTimeColumn(this._colName,
      {bool nullable: false,
      bool primary: false,
      Unique unique: const Unique.Not()})
      : _nullable = nullable,
        _primary = primary,
        _unique = unique;

  CreateDateTimeColumn nullable() {
    _nullable = true;
    return this;
  }

  String get colName => _colName;

  bool get isNullable => _nullable;

  bool get isPrimaryKey => _primary;

  Foreign get foreignKey => null;

  Unique get unique => _unique;

  String toSql() => _colName + ' TIMESTAMP' + (!_nullable ? ' NOT NULL' : '');
}

class CreateStrColumn extends CreateColumn<String> {
  bool _nullable = false;

  String _colName;

  bool _primary = false;

  int _length = 20;

  Foreign _foreignKey;

  Unique _unique;

  CreateStrColumn(this._colName,
      {bool nullable: false,
      bool primary: false,
      int length: 20,
      Foreign foreignKey,
      Unique unique: const Unique.Not()})
      : _nullable = nullable,
        _length = length,
        _primary = primary,
        _foreignKey = foreignKey,
        _unique = unique;

  CreateStrColumn.Primary(this._colName, {int length: 20, Foreign foreignKey})
      : _primary = true,
        _length = length,
        _foreignKey = foreignKey;

  CreateStrColumn nullable() {
    _nullable = true;
    return this;
  }

  CreateStrColumn foreign(String table, {String refCol}) {
    if (_foreignKey != null) throw new Exception("Foreign key alread set!");
    if (refCol == null) refCol = _colName;
    _foreignKey = new Foreign(table, refCol);
    return this;
  }

  String get colName => _colName;

  bool get isNullable => _nullable;

  bool get isPrimaryKey => _primary;

  int get length => length;

  Foreign get foreignKey => _foreignKey;

  Unique get unique => _unique;

  String toSql() =>
      _colName + ' VARCHAR(${_length})' + (!_nullable ? ' NOT NULL' : '');
}

class CreateStatement implements Statement {
  String _tableName;

  final Map<String, CreateColumn> _columns = {};

  bool _ifNotExists = false;

  CreateStatement();

  CreateStatement named(String tableName) {
    _tableName = tableName;
    return this;
  }

  CreateStatement ifNotExists([bool create = true]) {
    _ifNotExists = create;
    return this;
  }

  CreateStatement addInt(String colName,
      {bool autoIncrement: false,
      bool primary: false,
      String foreignTable,
      String foreignCol,
      Unique unique: const Unique.Not()}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateIntColumn(colName,
        autoIncrement: autoIncrement,
        primary: primary,
        foreignKey: foreign,
        unique: unique);
    return this;
  }

  CreateStatement addNullInt(String colName,
      {bool primary: false,
      Foreign foreign,
      String foreignTable,
      String foreignCol,
      Unique unique: const Unique.Not()}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateIntColumn(colName,
        nullable: true, primary: primary, foreignKey: foreign, unique: unique);
    return this;
  }

  CreateStatement addDouble(String colName,
      {bool autoIncrement: false,
      bool primary: false,
      String foreignTable,
      String foreignCol,
      Unique unique: const Unique.Not()}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateDoubleColumn(colName,
        autoIncrement: autoIncrement,
        primary: primary,
        foreignKey: foreign,
        unique: unique);
    return this;
  }

  CreateStatement addNullDouble(String colName,
      {bool primary: false,
      Foreign foreign,
      String foreignTable,
      String foreignCol,
      Unique unique: const Unique.Not()}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateDoubleColumn(colName,
        nullable: true, primary: primary, foreignKey: foreign, unique: unique);
    return this;
  }

  CreateStatement addBool(String colName, {Unique unique: const Unique.Not()}) {
    _columns[colName] = new CreateBoolColumn(colName, unique: unique);
    return this;
  }

  CreateStatement addNullBool(String colName,
      {Unique unique: const Unique.Not()}) {
    _columns[colName] = new CreateBoolColumn(colName).nullable();
    return this;
  }

  CreateStatement addDateTime(String colName,
      {Unique unique: const Unique.Not()}) {
    _columns[colName] = new CreateDateTimeColumn(colName, unique: unique);
    return this;
  }

  CreateStatement addNullDateTime(String colName,
      {Unique unique: const Unique.Not()}) {
    _columns[colName] =
        new CreateBoolColumn(colName, unique: unique).nullable();
    return this;
  }

  CreateStatement addStr(String colName,
      {int length: 20,
      bool primary: false,
      String foreignTable,
      String foreignCol,
      Unique unique: const Unique.Not()}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateStrColumn(colName,
        length: length, primary: primary, foreignKey: foreign, unique: unique);
    return this;
  }

  CreateStatement addNullStr(String colName,
      {int length: 20,
      bool primary: false,
      String foreignTable,
      String foreignCol,
      Unique unique: const Unique.Not()}) {
    Foreign foreign;
    if (foreignTable != null) {
      foreign =
          Foreign.make(foreignTable, foreignCol != null ? foreignCol : colName);
    }
    _columns[colName] = new CreateStrColumn(colName,
            length: length,
            primary: primary,
            foreignKey: foreign,
            unique: unique)
        .nullable();
    return this;
  }

  CreateStatement addColumn(CreateColumn column) {
    _columns[column.colName] = column;
    return this;
  }

  CreateStatement addColumns(List<CreateColumn> columns) {
    for (CreateColumn col in columns) {
      _columns[col.colName] = col;
    }
    return this;
  }

  String toSql() {
    StringBuffer sb = new StringBuffer();

    sb.write('CREATE TABLE');

    if (_ifNotExists) sb.write(' IF NOT EXISTS');

    sb.write(' $_tableName (');

    sb.write(_columns.values.map((CreateColumn col) => col.toSql()).join(', '));

    final List<CreateColumn> primaries =
        _columns.values.where((CreateColumn col) => col.isPrimaryKey).toList();
    if (primaries.length != 0) {
      sb.write(', PRIMARY KEY (');
      sb.write(primaries.map((CreateColumn col) => col.colName).join(','));
      sb.write(')');
    }

    {
      final y = <CreateColumn>[];
      final x = <String, List<CreateColumn>>{};
      for (CreateColumn col in _columns.values) {
        if (col.foreignKey != null) {
          sb.write(', FOREIGN KEY (${col.colName}) REFERENCES ${col.foreignKey
                  .table}(${col.foreignKey.col})');
        }

        if (col.unique.unique) {
          if (col.unique.group == null) {
            y.add(col);
          } else {
            x[col.unique.group] = (x[col.unique.group] ?? <CreateColumn>[])
              ..add(col);
          }
        }
      }

      for (CreateColumn col in y) {
        sb.write(', UNIQUE(${col.colName})');
      }

      for (String group in x.keys) {
        final String str =
            x[group].map((CreateColumn col) => col.colName).join(', ');
        sb.write(', UNIQUE($str)');
      }
    }

    sb.write(')');

    return sb.toString();
  }
}
